from pm4pymdl.algo.mvp.get_logs_and_replay.versions import inductive_and_tr

INDUCTIVE_AND_TR = "inductive_and_tr"

VERSIONS = {INDUCTIVE_AND_TR: inductive_and_tr.apply}

def apply(df, variant=INDUCTIVE_AND_TR, parameters=None):
    return VERSIONS[variant](df, parameters=parameters)



import pm4py
from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl, clean_frequency, clean_arc_frequency
from pm4pymdl.algo.mvp.projection import factory
from pm4py.algo.discovery.alpha import factory as alpha_miner
from pm4py.algo.discovery.inductive import factory as inductive_miner
from pm4py.algo.conformance.tokenreplay import factory as tr_factory
from pm4py.visualization.petrinet.util import performance_map
from pm4py.algo.filtering.log.variants import variants_filter as variants_module
from pm4py.algo.filtering.log.paths import paths_filter
from pm4py.algo.filtering.log.variants import variants_filter
from pm4py.algo.filtering.log.attributes import attributes_filter
from pm4py.objects.petri.petrinet import PetriNet, Marking
from pm4py.objects.petri.utils import add_arc_from_to
from pm4py.objects.petri.utils import remove_place, remove_transition
from copy import deepcopy
import uuid


PARAM_ACTIVITY_KEY = pm4py.util.constants.PARAMETER_CONSTANT_ACTIVITY_KEY


def reduce_petri_net(net):
    transes = set([x for x in net.transitions if x.label is None])
    places = list(net.places)
    i = 0
    while i < len(places):
        place = places[i]
        source_transes = set([x.source for x in place.in_arcs])
        target_transes = set([x.target for x in place.out_arcs])
        if len(source_transes) == 1 and len(target_transes) == 1:
            if source_transes.issubset(transes) and target_transes.issubset(transes):
                source_trans = list(source_transes)[0]
                target_trans = list(target_transes)[0]
                if len(target_trans.out_arcs) == 1:
                    target_place = list(target_trans.out_arcs)[0].target
                    add_arc_from_to(source_trans, target_place, net)
                    remove_place(net, place)
                    remove_transition(net, target_trans)
                    places = list(net.places)
                    continue
                    #print(source_trans, target_trans, target_place)
        i = i + 1

    return net


def apply(df, parameters=None):
    if parameters is None:
        parameters = {}

    allowed_activities = parameters["allowed_activities"] if "allowed_activities" in parameters else None

    if df.type == "succint":
        df = succint_mdl_to_exploded_mdl.apply(df)
        df.type = "exploded"

    min_node_freq = parameters["min_node_freq"] if "min_node_freq" in parameters else 0
    min_edge_freq = parameters["min_edge_freq"] if "min_edge_freq" in parameters else 0

    df = clean_frequency.apply(df, min_node_freq)
    df = clean_arc_frequency.apply(df, min_edge_freq)

    persps = [x for x in df.columns if not x.startswith("event_")]

    ret = {}
    ret["nets"] = {}
    ret["act_count"] = {}
    ret["replay"] = {}
    ret["group_size_hist"] = {}
    ret["act_count_replay"] = {}
    ret["group_size_hist_replay"] = {}
    ret["aligned_traces"] = {}
    ret["place_fitness_per_trace"] = {}
    ret["aggregated_statistics_frequency"] = {}
    ret["aggregated_statistics_performance_min"] = {}
    ret["aggregated_statistics_performance_max"] = {}
    ret["aggregated_statistics_performance_median"] = {}
    ret["aggregated_statistics_performance_mean"] = {}

    for persp in persps:
        print(persp,"getting log")
        log = factory.apply(df, persp, parameters=parameters)
        print(len(log))


        if allowed_activities is not None:
            if persp not in allowed_activities:
                continue
            filtered_log = attributes_filter.apply_events(log, allowed_activities[persp])
        else:
            filtered_log = log

        # filtered_log = variants_filter.apply_auto_filter(deepcopy(filtered_log), parameters={"decreasingFactor": 0.5})

        print(len(log))
        print(persp,"got log")

        net, im, fm = inductive_miner.apply(filtered_log)

        """if persp == "items":
            trans_map = {t.label:t for t in net.transitions}
            source_place_it = list(trans_map["item out of stock"].in_arcs)[0].source
            target_place_re = list(trans_map["reorder item"].out_arcs)[0].target
            skip_trans_1 = PetriNet.Transition(str(uuid.uuid4()), None)
            net.transitions.add(skip_trans_1)
            add_arc_from_to(source_place_it, skip_trans_1, net)
            add_arc_from_to(skip_trans_1, target_place_re, net)"""

        net = reduce_petri_net(net)
        #net, im, fm = alpha_miner.apply(filtered_log)
        print(persp,"got model")

        activ_count = factory.apply(df, persp, variant="activity_occurrence", parameters=parameters)
        print(persp,"got activ_count")

        variants_idx = variants_module.get_variants_from_log_trace_idx(log)
        #variants = variants_module.convert_variants_trace_idx_to_trace_obj(log, variants_idx)
        #parameters_tr = {PARAM_ACTIVITY_KEY: "concept:name", "variants": variants}

        print(persp,"got variants")


        aligned_traces, place_fitness_per_trace, transition_fitness_per_trace, notexisting_activities_in_model = tr_factory.apply(log, net, im, fm, parameters={"enable_place_fitness": True, "disable_variants": True})

        print(persp,"done tbr")

        element_statistics = performance_map.single_element_statistics(log, net, im,
                                                                       aligned_traces, variants_idx)

        print(persp,"done element_statistics")

        aggregated_statistics = performance_map.aggregate_statistics(element_statistics)

        print(persp,"done aggregated_statistics")

        element_statistics_performance = performance_map.single_element_statistics(log, net, im,
                                                                       aligned_traces, variants_idx)

        print(persp,"done element_statistics_performance")

        aggregated_statistics_performance_min = performance_map.aggregate_statistics(element_statistics_performance, measure="performance", aggregation_measure="min")
        aggregated_statistics_performance_max = performance_map.aggregate_statistics(element_statistics_performance, measure="performance", aggregation_measure="max")
        aggregated_statistics_performance_median = performance_map.aggregate_statistics(element_statistics_performance, measure="performance", aggregation_measure="median")
        aggregated_statistics_performance_mean = performance_map.aggregate_statistics(element_statistics_performance, measure="performance", aggregation_measure="mean")

        print(persp,"done aggregated_statistics_performance")

        group_size_hist = factory.apply(df, persp, variant="group_size_hist", parameters=parameters)

        print(persp,"done group_size_hist")

        occurrences = {}
        for trans in transition_fitness_per_trace:
            occurrences[trans.label] = set()
            for trace in transition_fitness_per_trace[trans]["fit_traces"]:
                if not trace in transition_fitness_per_trace[trans]["underfed_traces"]:
                    case_id = trace.attributes["concept:name"]
                    for event in trace:
                        if event["concept:name"] == trans.label:
                            occurrences[trans.label].add((case_id, event["event_id"]))
            #print(transition_fitness_per_trace[trans])


        len_different_ids = {}
        for act in occurrences:
            len_different_ids[act] = len(set(x[1] for x in occurrences[act]))

        eid_acti_count = {}
        for act in occurrences:
            eid_acti_count[act] = {}
            for x in occurrences[act]:
                if not x[0] in eid_acti_count:
                    eid_acti_count[act][x[0]] = 0
                eid_acti_count[act][x[0]] = eid_acti_count[act][x[0]] + 1
            eid_acti_count[act] = sorted(list(eid_acti_count[act].values()))

        ret["nets"][persp] = [net, im, fm]
        ret["act_count"][persp] = activ_count
        ret["aligned_traces"][persp] = aligned_traces
        ret["place_fitness_per_trace"][persp] = place_fitness_per_trace
        ret["aggregated_statistics_frequency"][persp] = aggregated_statistics
        ret["aggregated_statistics_performance_min"][persp] = aggregated_statistics_performance_min
        ret["aggregated_statistics_performance_max"][persp] = aggregated_statistics_performance_max
        ret["aggregated_statistics_performance_median"][persp] = aggregated_statistics_performance_median
        ret["aggregated_statistics_performance_mean"][persp] = aggregated_statistics_performance_mean

        ret["replay"][persp] = aggregated_statistics
        ret["group_size_hist"][persp] = group_size_hist
        ret["act_count_replay"][persp] = len_different_ids
        ret["group_size_hist_replay"][persp] = eid_acti_count

        #print(ret["act_count"])
        #print(ret["act_count_replay"])


    return ret



import math
import random
import string
import time
import numpy as np

import pandas as pd


def generate_random_string(N):
    """
    Generate a random string

    Parameters
    ------------
    N
        length of the random string

    Returns
    ------------
    stru
        Random string
    """
    return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(N))


def get_time(timestamp):
    """
    Get a string representing the timestamp

    Parameters
    ------------
    timestamp
        Numeric timestamp

    Returns
    ------------
    stru
        String timestamp
    """
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(timestamp)).replace(" ", "T") + ".000"


def generate_mvp_log(n_events=1000, n_classes=None, n_activities=None, n_objects_per_class=None):
    """
    Generate a MVP model

    Parameters
    -------------
    n_events
        Number of events
    n_classes
        Number of classes
    n_activities
        Number of activities
    n_objects_per_class
        Number of objects per class

    Returns
    -------------
    log
        MVP log
    """
    if n_classes is None:
        # default number of classes
        n_classes = max(10, math.floor(n_events / 1000))

    if n_objects_per_class is None:
        # default number of objects per class
        n_objects_per_class = math.floor(n_events / (n_classes))

    if n_activities is None:
        # default number of activities
        n_activities = max(10, math.floor(n_events / 10000))

    print("generating model with events=",n_events, "objects=",(n_objects_per_class*n_classes)," classes=",n_classes," activities=",n_activities)

    this_timestamp = 0

    all_activities = []

    for i in range(n_activities):
        all_activities.append("activity_"+str(i))

    classes = []
    objects_per_class = {}

    columns = {"event_id": [], "event_activity": [], "event_timestamp": []}

    for i in range(n_classes):
        this_class = generate_random_string(20)
        classes.append(this_class)
        columns[this_class] = []
        objects_per_class[this_class] = []
        for j in range(n_objects_per_class):
            objects_per_class[this_class].append(this_class + "_" + str(j))

    for i in range(n_events):
        event_name = "event_" + str(i)

        this_timestamp = this_timestamp + 1
        timestamp = get_time(this_timestamp)

        activity = all_activities[random.randrange(0, len(all_activities))]

        prev_idx = len(columns["event_id"])

        #print(classes)

        for j in range(n_classes):
            columns["event_id"].append(event_name)
            columns["event_activity"].append(activity)
            columns["event_timestamp"].append(timestamp)
            for col in classes:
                columns[col].append(np.nan)

        for j in range(n_classes):
            classe = classes[j]
            chosen_object_value = objects_per_class[classe][random.randrange(0,len(objects_per_class[classe]))]

            columns[classe][prev_idx + j] = chosen_object_value
        if i % 10000 == 0 and i > 0:
            print(i,"of",n_events)
    dataframe = pd.DataFrame(data=columns)
    return dataframe



import sqlite3
from datetime import datetime

import pandas as pd


def apply(file_path, parameters=None):
    """
    Import an OpenSLEX model

    Parameters
    -------------
    file_path
        File path
    parameters
        Possible parameters of the algorithm

    Returns
    -------------
    dataframe
        Dataframe
    """
    if parameters is None:
        parameters = {}

    conn = sqlite3.connect(file_path)
    classes = {}
    object_classes = {}
    object_versions = {}
    activities = {}
    activities_instances = {}
    events = {}
    event_to_objv = {}
    relations = {}

    c = conn.cursor()
    c.execute("SELECT id, name FROM class")
    for result in c.fetchall():
        classes[result[0]] = result[1]
    c.execute("SELECT id, class_id FROM object")
    for result in c.fetchall():
        object_classes[result[0]] = result[1]
    c.execute("SELECT id, name FROM activity")
    for result in c.fetchall():
        activities[result[0]] = result[1]
    c.execute("SELECT id, name FROM activity")
    for result in c.fetchall():
        activities[result[0]] = result[1]
    c.execute("SELECT id, activity_id FROM activity_instance")
    for result in c.fetchall():
        activities_instances[result[0]] = result[1]
    c.execute("SELECT id, object_id FROM object_version")
    for result in c.fetchall():
        object_versions[result[0]] = result[1]
    c.execute("SELECT id, activity_instance_id, timestamp FROM event")
    for result in c.fetchall():
        events[result[0]] = (result[1], result[2])
    c.execute("SELECT event_id, object_version_id FROM event_to_object_version")
    for result in c.fetchall():
        if not result[0] in event_to_objv:
            event_to_objv[result[0]] = []
        event_to_objv[result[0]].append(result[1])
    c.execute("SELECT id, source_object_version_id, target_object_version_id FROM relation")
    for result in c.fetchall():
        if not result[0] in relations:
            relations[result[0]] = []
        relations[result[0]].append(result[1])

    stream = []
    for event in events:
        if event in event_to_objv:
            activity = activities[activities_instances[events[event][0]]]
            timestamp = datetime.utcfromtimestamp(events[event][1] / 1000.0)
            for source_objv in event_to_objv[event]:
                source_obj_id = object_versions[source_objv]
                source_obj_class = classes[object_classes[source_obj_id]]
                stream.append({"event_id": event, "event_activity": activity, "event_timestamp": timestamp,
                               source_obj_class: source_obj_id})
                if source_objv in relations:
                    for target_objv in relations[source_objv]:
                        target_obj_id = object_versions[target_objv]
                        target_obj_class = classes[object_classes[target_obj_id]]
                        stream.append({"event_id": event, "event_activity": activity, "event_timestamp": timestamp,
                                       target_obj_class: target_obj_id})
    dataframe = pd.DataFrame.from_dict(stream)
    dataframe = dataframe.sort_values("event_timestamp")

    dataframe.type = "exploded"

    return dataframe



from pm4pymdl.objects.openslex.importer.versions import classic

CLASSIC = "classic"

VERSIONS = {CLASSIC: classic.apply}


def apply(file_path, parameters=None, variant=CLASSIC):
    """
    Import an OpenSLEX model

    Parameters
    -------------
    file_path
        File path
    parameters
        Possible parameters of the algorithm
    variant
        Variant of the algorithm, possible values: class

    Returns
    -------------
    dataframe
        Dataframe
    """
    return VERSIONS[variant](file_path, parameters=parameters)



import pandas as pd


def apply(df, model, parameters=None):
    if parameters is None:
        parameters = {}

    perspectives = [x for x in df.columns if not x.startswith("event_")]

    ret = {}
    basic_columns = ["event_id", "event_activity"]

    for c in perspectives:
        proj_df = df[basic_columns + [c]].dropna()

        first_df = proj_df.groupby(c).first()

        merged_df = first_df.merge(proj_df, how='inner', left_on=c, right_on=c, suffixes=('', '_2'))
        merged_df = merged_df[merged_df["event_id"] != merged_df["event_id_2"]]
        merged_df["event_activity_merge"] = merged_df["event_activity"] + "@@" + merged_df["event_activity_2"]

        merged_df = merged_df.reset_index()

        ret[c] = merged_df

    return ret



import uuid
import tempfile
from graphviz import Digraph
from pm4py.objects.petri.petrinet import PetriNet
from statistics import median, mean

COLORS = ["#05B202", "#A13CCD", "#BA0D39", "#39F6C0", "#E90638", "#07B423", "#306A8A", "#678225", "#2742FE", "#4C9A75",
          "#4C36E9", "#7DB022", "#EDAC54", "#EAC439", "#EAC439", "#1A9C45", "#8A51C4", "#496A63", "#FB9543", "#2B49DD",
          "#13ADA5", "#2DD8C1", "#2E53D7", "#EF9B77", "#06924F", "#AC2C4D", "#82193F", "#0140D3"]


def apply(obj, parameters=None):
    if parameters is None:
        parameters = {}

    image_format = "png"
    if "format" in parameters:
        image_format = parameters["format"]

    filename = tempfile.NamedTemporaryFile(suffix='.gv').name
    g = Digraph("", filename=filename, engine='dot', graph_attr={'bgcolor': 'transparent'})

    nets = obj["nets"]
    replay = obj["replay"]
    act_count = obj["act_count"]
    group_size = obj["group_size_hist_replay"]
    aligned_traces = obj["aligned_traces"]
    place_fitness_per_trace_persp = obj["place_fitness_per_trace"]
    group_size_hist_persp = obj["group_size_hist"]
    aggregated_statistics_frequency = obj["aggregated_statistics_frequency"]
    aggregated_statistics_performance_min = obj["aggregated_statistics_performance_min"]
    aggregated_statistics_performance_max = obj["aggregated_statistics_performance_max"]
    aggregated_statistics_performance_median = obj["aggregated_statistics_performance_median"]
    aggregated_statistics_performance_mean = obj["aggregated_statistics_performance_mean"]
    aligned_traces = obj["aligned_traces"]

    all_objs = {}
    trans_names = {}

    for index, persp in enumerate(nets):
        orig_act_count = {}
        orig_arc_count = {}

        color = COLORS[index % len(COLORS)]
        net, im, fm = nets[persp]
        rr = replay[persp]
        ac = act_count[persp]
        gs = group_size[persp]

        ag_freq = aggregated_statistics_frequency[persp]
        ag_perf_min = aggregated_statistics_performance_min[persp]
        ag_perf_max = aggregated_statistics_performance_max[persp]
        ag_perf_median = aggregated_statistics_performance_median[persp]
        ag_perf_mean = aggregated_statistics_performance_mean[persp]
        al_trac = aligned_traces[persp]
        place_fitness_per_trace = place_fitness_per_trace_persp[persp]
        group_size_hist = group_size_hist_persp[persp]

        for pl in place_fitness_per_trace:
            if place_fitness_per_trace[pl]["c"] + place_fitness_per_trace[pl]["r"] > place_fitness_per_trace[pl]["p"] + \
                    place_fitness_per_trace[pl]["m"]:
                """place_fitness_per_trace[pl]["m"] += place_fitness_per_trace[pl]["c"] + place_fitness_per_trace[pl][
                    "r"] - \
                                                    place_fitness_per_trace[pl]["p"]"""
                place_fitness_per_trace[pl]["c"] = place_fitness_per_trace[pl]["p"] + place_fitness_per_trace[pl]["m"] - \
                                                   place_fitness_per_trace[pl]["r"]
            elif place_fitness_per_trace[pl]["c"] + place_fitness_per_trace[pl]["r"] < place_fitness_per_trace[pl][
                "p"] + \
                    place_fitness_per_trace[pl]["m"]:
                """place_fitness_per_trace[pl]["r"] += place_fitness_per_trace[pl]["p"] + place_fitness_per_trace[pl][
                    "m"] - \
                                                    place_fitness_per_trace[pl]["c"]"""
                place_fitness_per_trace[pl]["p"] = place_fitness_per_trace[pl]["c"] + place_fitness_per_trace[pl]["r"] - \
                                                   place_fitness_per_trace[pl]["m"]

        # print("ag_freq",ag_freq)
        # print("ag_perf_min",ag_perf_min)
        # print("ag_perf_max",ag_perf_max)
        # print("ag_perf_median",ag_perf_median)
        # print("ag_perf_mean",ag_perf_mean)
        # print("al_trac",al_trac)
        # print("pl_fitness",place_fitness_per_trace)

        # input()

        for pl in net.places:
            this_uuid = str(uuid.uuid4())
            pl_str = "p=%d r=%d\\nc=%d m=%d" % (
                place_fitness_per_trace[pl]["p"], place_fitness_per_trace[pl]["r"], place_fitness_per_trace[pl]["c"],
                place_fitness_per_trace[pl]["m"])
            is_double_circled = False
            for arc in pl.in_arcs:
                if arc.source.label in group_size_hist:
                    if len(group_size_hist[arc.source.label]) != sum(group_size_hist[arc.source.label]):
                        is_double_circled = True
            for arc in pl.out_arcs:
                if arc.target.label in group_size_hist:
                    if len(group_size_hist[arc.target.label]) != sum(group_size_hist[arc.target.label]):
                        is_double_circled = True
            if is_double_circled:
                if pl in im:
                    g.node(this_uuid, pl_str, shape="doublecircle", style="filled", color=color, fillcolor=color,
                           fontsize="13.0", labelfontsize="13.0", penwidth="3.0")
                elif pl in fm:
                    g.node(this_uuid, pl_str, shape="doublecircle", style="filled", color=color, fillcolor=color,
                           fontsize="13.0", labelfontsize="13.0", penwidth="3.0")
                else:
                    g.node(this_uuid, pl_str, shape="doublecircle", color=color, fontsize="13.0", labelfontsize="13.0", penwidth="3.0")
            else:
                if pl in im:
                    g.node(this_uuid, pl_str, shape="circle", style="filled", fillcolor=color, color=color,
                           fontsize="13.0", labelfontsize="13.0")
                elif pl in fm:
                    g.node(this_uuid, pl_str, shape="circle", style="filled", color=color, fillcolor=color,
                           fontsize="13.0", labelfontsize="13.0")
                else:
                    g.node(this_uuid, pl_str, shape="circle", color=color, fontsize="13.0", labelfontsize="13.0")
            all_objs[pl] = this_uuid

        for tr in net.transitions:
            this_uuid = str(uuid.uuid4())
            if tr.label is None:
                g.node(this_uuid, "", shape="box", fillcolor="#000000", style="filled")
                all_objs[tr] = this_uuid
            elif tr.label not in trans_names:
                count = rr[tr]["label"].split("(")[-1].split("]")[0]
                orig_act_count[tr] = count
                this_act_count = "%d" % (float(ac[tr.label]))
                g.node(this_uuid, tr.label + " \\n" + this_act_count + "", shape="box", fontsize="36.0",
                       labelfontsize="36.0")
                trans_names[tr.label] = this_uuid
                all_objs[tr] = this_uuid
            else:
                all_objs[tr] = trans_names[tr.label]

        for arc in net.arcs:
            this_uuid = str(uuid.uuid4())

            source_node = arc.source
            target_node = arc.target
            to_double_arc = False

            if arc in rr:

                arc_count = float(rr[arc]['label'])

                if arc in ag_perf_min:
                    perf_str = "\\nperf: median=%s mean=%s" % (
                        ag_perf_median[arc]['label'],
                        ag_perf_mean[arc]['label'])
                else:
                    perf_str = ""

                if type(source_node) is PetriNet.Place:
                    if True or (target_node.label is not None and ac[target_node.label] > 0):
                        # if ac[target_node.label] > 0:
                        if target_node.label in group_size_hist and max(group_size_hist[target_node.label]) > 1:
                            pre = "freq: median=%d mean=%.2f\\neve=%d uniqobj=%d\\n" % (
                                median(group_size_hist[target_node.label]), mean(group_size_hist[target_node.label]),
                                len(group_size_hist[target_node.label]), sum(group_size_hist[target_node.label]))
                        else:
                            # pre="freq: %d\\n" % len(group_size_hist[target_node.label])
                            pre = "uniqobj=" + str(int(arc_count))
                        if target_node.label in group_size_hist and sum(group_size_hist[target_node.label]) != len(
                                group_size_hist[target_node.label]):
                            to_double_arc = True

                        ratio = pre + perf_str + ""
                    else:
                        # ratio = "1.00" + perf_str + ""
                        ratio = "freq: " + str(int(arc_count))
                else:
                    if True or (source_node.label is not None and ac[source_node.label] > 0):
                        # if ac[source_node.label] > 0:
                        if source_node.label in group_size_hist and max(group_size_hist[source_node.label]) > 1:
                            pre = "freq: median=%d mean=%.2f\\neve=%d uniqobj=%d\\n" % (
                                median(group_size_hist[source_node.label]), mean(group_size_hist[source_node.label]),
                                len(group_size_hist[source_node.label]), sum(group_size_hist[source_node.label]))
                        else:
                            # pre="freq: %d\\n" % len(group_size_hist[source_node.label])
                            pre = "uniqobj=" + str(int(arc_count))
                        if source_node.label in group_size_hist and sum(group_size_hist[source_node.label]) != len(
                                group_size_hist[source_node.label]):
                            to_double_arc = True

                        ratio = pre + perf_str + ""
                    else:
                        # ratio = "1.00" + perf_str + ""
                        ratio = "freq: " + str(int(arc_count))

                if to_double_arc:
                    g.edge(all_objs[source_node], all_objs[target_node], label=ratio, penwidth=str(3.0*float(rr[arc]['penwidth'])),
                           color=color + ":white:" + color, fontsize="13.0")
                else:
                    g.edge(all_objs[source_node], all_objs[target_node], label=ratio, penwidth=str(1.2*float(rr[arc]['penwidth'])),
                           color=color, fontsize="13.0")

                all_objs[arc] = this_uuid
            else:
                g.edge(all_objs[source_node], all_objs[target_node], label="", color=color, fontsize="11.0")

    g.attr(overlap='false')
    g.attr(fontsize='11')

    g.format = image_format

    return g



from pm4pymdl.visualization.petrinet.versions import from_log_and_replay
from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave


FROM_LOG_AND_REPLAY = "from_log_and_replay"

VERSIONS = {FROM_LOG_AND_REPLAY: from_log_and_replay.apply}

def apply(obj, variant=FROM_LOG_AND_REPLAY, parameters=None):
    return VERSIONS[variant](obj, parameters=parameters)


def save(gviz, output_file_path):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path)


def view(gviz):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    return gview.view(gviz)



from pm4pymdl.algo.mvp.gen_framework.models import model1
from pm4pymdl.algo.mvp.gen_framework.models import model2
from pm4pymdl.algo.mvp.gen_framework.models import model3


MODEL1 = "model1"
MODEL2 = "model2"
MODEL3 = "model3"

VERSIONS = {MODEL1: model1.MVPModel1, MODEL2: model2.MVPModel2, MODEL3: model3.MVPModel3}


def apply(df, variant=MODEL1, parameters=None):
    model = VERSIONS[variant](df, parameters=parameters)
    if variant == MODEL2:
        model.calculate_map()
    return model




from pm4py.objects.log.importer.parquet import factory as parquet_importer
import pandas as pd

def apply(file_path, parameters=None):
    if parameters is None:
        parameters = {}

    sep = parameters["sep"] if "sep" in parameters else ","
    quotechar = parameters["quotechar"] if "quotechar" in parameters else "\""

    if file_path.endswith(".csv") or file_path.endswith(".mdl"):
        df = pd.read_csv(file_path, sep=sep, quotechar=quotechar)
        df.type = "succint"
        df["event_timestamp"] = pd.to_datetime(df["event_timestamp"])
        return df
    elif file_path.endswith(".parquet"):
        df = parquet_importer.apply(file_path)
        df.type = "exploded"
        df["event_timestamp"] = pd.to_datetime(df["event_timestamp"])
        return df



from pm4pymdl.algo.mvp.gen_framework.rel_activities import classic

CLASSIC = "classic"

VERSIONS = {CLASSIC: classic.apply}


def apply(df, model, rel_ev_dict, variant=CLASSIC, parameters=None):
    if parameters is None:
        parameters = {}

    return VERSIONS[variant](df, model, rel_ev_dict, parameters=parameters)



from pm4pymdl.algo.mvp.gen_framework.models import factory as model_factory
from pm4pymdl.algo.mvp.gen_framework.rel_events import factory as rel_ev_factory
from pm4pymdl.algo.mvp.gen_framework.rel_activities import factory as rel_act_factory
from pm4pymdl.algo.mvp.gen_framework.node_freq import factory as node_freq_factory
from pm4pymdl.algo.mvp.gen_framework.edge_freq import factory as edge_freq_factory
from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl, clean_objtypes

MODEL1 = model_factory.MODEL1
MODEL2 = model_factory.MODEL2
MODEL3 = model_factory.MODEL3
REL_DFG = rel_ev_factory.REL_DFG
BEING_PRODUCED = rel_ev_factory.BEING_PRODUCED
TYPE1 = node_freq_factory.TYPE1
TYPE11 = edge_freq_factory.TYPE11
TYPE12 = edge_freq_factory.TYPE12

COLORS = {REL_DFG: "#000000"}

def apply(df, model_type_variant=MODEL1, rel_ev_variant=REL_DFG, node_freq_variant=TYPE1, edge_freq_variant=TYPE11, parameters=None):
    if parameters is None:
        parameters = {}

    conversion_needed = False

    try:
        if df.type == "succint":
            conversion_needed = True
    except:
        pass

    if conversion_needed:
        df = succint_mdl_to_exploded_mdl.apply(df)

    df = clean_objtypes.perfom_cleaning(df, parameters=parameters)

    model = model_factory.apply(df, variant=model_type_variant)
    rel_ev = rel_ev_factory.apply(df, model, variant=rel_ev_variant)
    rel_act = rel_act_factory.apply(df, model, rel_ev)
    node_freq = node_freq_factory.apply(df, model, rel_ev, rel_act, variant=node_freq_variant)
    edge_freq = edge_freq_factory.apply(df, model, rel_ev, rel_act, variant=edge_freq_variant)

    model.set_rel_ev(rel_ev)
    model.set_rel_act(rel_act)
    model.set_node_freq(node_freq)
    model.set_edge_freq(edge_freq)

    return model



class BasicModelStructure(object):
    def __init__(self, df, parameters=None):
        if parameters is None:
            parameters = {}

        self.df = df
        self.rel_ev = None
        self.rel_act = None
        self.node_freq = None
        self.edge_freq = None

    def get_rel_ev(self):
        return self.rel_ev

    def set_rel_ev(self, rel_ev):
        self.rel_ev = rel_ev

    def get_rel_act(self):
        return self.rel_act

    def set_rel_act(self, rel_act):
        self.rel_act = rel_act

    def get_node_freq(self):
        return self.node_freq

    def set_node_freq(self, node_freq):
        self.node_freq = node_freq

    def get_edge_freq(self):
        return self.edge_freq

    def set_edge_freq(self, edge_freq):
        self.edge_freq = edge_freq

    def get_perspectives(self):
        if self.rel_ev is not None:
            return set(self.rel_ev.keys())
        return set()

    def get_all_activities(self):
        all_activities = set()
        if self.node_freq is not None:
            for el in self.node_freq:
                if type(self.node_freq[el]) is str:
                    all_activities = all_activities.union(self.node_freq[el])
                else:
                    all_activities = set(self.node_freq.keys())
        return all_activities




import pandas as pd


def apply(df, model, parameters=None):
    if parameters is None:
        parameters = {}

    ret = {}
    columns = [x for x in df.columns if not x.startswith("event_")]
    basic_columns = ["event_id", "event_activity"]
    for col in columns:
        proj_df = df[basic_columns + [col]].dropna()
        proj_df["@@index"] = proj_df.index
        proj_df = proj_df.sort_values([col, "@@index"])
        proj_df_shift = proj_df.shift(-1)
        proj_df_shift.columns = [str(c) + '_2' for c in proj_df_shift.columns]
        concat_df = pd.concat([proj_df, proj_df_shift], axis=1)
        concat_df = concat_df[concat_df[col] == concat_df[col+"_2"]]

        concat_df["event_activity_merge"] = concat_df["event_activity"] + "@@" + concat_df["event_activity_2"]

        ret[col] = concat_df

    return ret



import pandas as pd


def apply(df, model, parameters=None):
    if parameters is None:
        parameters = {}

    perspectives = [x for x in df.columns if not x.startswith("event_")]

    ret = {}
    basic_columns = ["event_id", "event_activity"]

    for c in perspectives:
        proj_df = df[basic_columns + [c]].dropna()
        proj_df_2 = proj_df.copy()

        merged_df = proj_df.merge(proj_df_2, how='inner', left_on=[c, "event_activity"], right_on=[c, "event_activity"], suffixes=('','_2'))
        #merged_df = merged_df[merged_df["event_id"] == merged_df["event_id_2"]]
        merged_df["@@eid_c_concat"] = merged_df["event_id"] + "@@@" + merged_df[c]

        g_m_df = merged_df["@@eid_c_concat"].groupby(merged_df["@@eid_c_concat"]).transform('size')

        merged_df = merged_df[g_m_df == 1]

        merged_df["event_activity_2"] = merged_df["event_activity"]
        merged_df["event_activity_merge"] = merged_df["event_activity"] + "@@" + merged_df["event_activity_2"]
        merged_df = merged_df.reset_index()

        ret[c] = merged_df

    return ret



from graphviz import Digraph
from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave
import uuid
import tempfile

def apply(model, parameters=None):
    if parameters is None:
        parameters = {}

    image_format = "png"
    if "format" in parameters:
        image_format = parameters["format"]

    filename = tempfile.NamedTemporaryFile(suffix='.gv').name
    g = Digraph("", filename=filename, engine='dot', graph_attr={'bgcolor': 'transparent'})

    activities = {}
    cluster_corr = {}
    all_objs = {}

    for p in model.model_inductive:
        net, im, fm = model.model_inductive[p]
        activities[p] = {}

        with g.subgraph(name='cluster_' + p) as c:
            cluster_corr[p] = c
            c.attr(style='filled')
            c.attr(color='lightgrey')
            c.node_attr.update(style='filled', color='white')
            c.attr(label='class: ' + p)

            for pl in net.places:
                this_uuid = str(uuid.uuid4())
                if pl in im:
                    c.node(this_uuid, "", shape="circle", fillcolor="#32CD32")
                elif pl in fm:
                    c.node(this_uuid, "", shape="circle", fillcolor="#FFA500")
                else:
                    c.node(this_uuid, "", shape="circle")
                all_objs[pl] = this_uuid

            for tr in net.transitions:
                this_uuid = str(uuid.uuid4())
                if tr.label is None:
                    c.node(this_uuid, "", shape="box", fillcolor="#000000", style="filled")
                else:
                    c.node(this_uuid, tr.label, shape="box")
                    activities[p][tr.label] = this_uuid
                all_objs[tr] = this_uuid

            for arc in net.arcs:
                c.edge(all_objs[arc.source], all_objs[arc.target])

    for scl in model.production_map:
        for tcl in model.production_map[scl]:
            for acti in model.production_map[scl][tcl]:
                st = activities[scl][acti]
                tt = activities[tcl][acti]

                this_uuid = str(uuid.uuid4())
                g.node(this_uuid, "", shape="box", fillcolor="#99FF99", style="filled")

                g.edge(st, this_uuid, dir="none", xlabel=model.production_map[scl][tcl][acti]["source_card"])
                g.edge(this_uuid, tt, dir="none", xlabel=model.production_map[scl][tcl][acti]["target_card"])

                for la_el in model.production_map[scl][tcl][acti]["linked"]:
                    la = la_el[0]
                    lac = la_el[1]

                    st2 = activities[scl][la]
                    tt2 = activities[tcl][la]

                    this_uuid = str(uuid.uuid4())
                    if str(lac) == "<=":
                        g.node(this_uuid, str(lac), shape="box", fillcolor="#AAEEEE", style="filled")
                    else:
                        g.node(this_uuid, str(lac), shape="box", fillcolor="#AAAAFF", style="filled")

                    g.edge(st2, this_uuid, dir="none")
                    g.edge(this_uuid, tt2, dir="none")

    g.attr(overlap='false')
    g.attr(fontsize='11')

    g.format = image_format

    return g


def apply(df, model, rel_ev, rel_act, parameters=None):
    if parameters is None:
        parameters = {}

    ret = {}

    for persp in rel_ev:
        df = rel_ev[persp]
        df = df[df["event_activity_merge"].isin(rel_act[persp])]
        df = df.groupby(["event_id", "event_id_2"]).first()

        r = df.groupby("event_activity_merge").size().to_dict()

        ret[persp] = r

    return ret



from pm4pymdl.objects.xoc.importer.versions import classic

CLASSIC = "classic"

VERSIONS = {CLASSIC: classic.apply}


def apply(file_path, parameters=None, variant=CLASSIC):
    """
    Apply the importing of a XOC file

    Parameters
    ------------
    file_path
        Path to the XOC file
    parameters
        Import parameters
    variant
        Variant of the algorithm, including: classic

    Returns
    ------------
    dataframe
        Dataframe
    """
    return VERSIONS[variant](file_path, parameters=parameters)



import pandas as pd

DEFAULT_ACTIVITY_KEY = "event_activity"


def apply(df, map, activity_key=DEFAULT_ACTIVITY_KEY, second_required=False, timestamp_required=True,
          merge_required=False, parameters=None):
    if parameters is None:
        parameters = {}

    df_list = []

    class_dest = {}

    for act in map:
        target = map[act]
        if target not in class_dest:
            class_dest[target] = []
        class_dest[target].append(act)

    for cl in class_dest:
        if cl in df.columns:
            cols_to_include0 = ["event_id", "event_activity", cl]
            if timestamp_required:
                cols_to_include0.append("event_timestamp")

            if not second_required:
                cols_to_include = cols_to_include0
            else:
                cols_to_include = [x for x in cols_to_include0] + [x + "_2" for x in cols_to_include0]

            if activity_key not in cols_to_include:
                cols_to_include.append(activity_key)
            if merge_required:
                cols_to_include.append("event_activity_merge")

            red_df = df[cols_to_include].dropna()
            red_df = red_df[red_df[activity_key].isin(class_dest[cl])]

            df_list.append(red_df)

    new_df = pd.concat(df_list).reset_index()

    return new_df



def get(dataframe, activity, parameters=None):
    if parameters is None:
        parameters = {}

    ret = {}
    red_df = dataframe[dataframe["event_activity"] == activity]
    red_df = red_df.dropna(axis=1, how="all")
    cols = [x for x in red_df.columns if not x.startswith("event_")]

    return cols
